/* Josh Pieper, (c) 2000 */

/* This file is distributed under the GPL, see file COPYING for details */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "gnut_lib.h"
#include "Gnut_List.h"
#include "Gnut_Hash.h"

int gnut_hash_foreach(Gnut_Hash *gh, int (*a)(void *, void *), void *data)
{
	int i;
	
	for (i=0;i<256;i++) if (gnut_list_foreach(gh->list[i],a,data)==-1) return -1;
	return 0;
}

Gnut_Hash * gnut_hash_new(uchar (*hash_func)(void *),
	int (*compare_func)(void *, void *))
{
	Gnut_Hash *gh;
	int i=0;
	
	gh=(Gnut_Hash *) xmalloc(sizeof(Gnut_Hash));
	for (i=0;i<256;i++) gh->list[i]=NULL;	
	gh->hash_func=hash_func;
	gh->compare_func=compare_func;
	
	return gh;
}

void gnut_hash_insert(Gnut_Hash *gh, void *data)
{
	int a;

	if (!gh) {
		g_debug(0,"gh=%p\n data=%p\n",gh,data);
		exit(0);
	}
	
	a=(*gh->hash_func)(data);
	g_debug(5,"hash returned: %i\n",a);

	gh->list[a]=gnut_list_prepend(gh->list[a],data);
}

void * gnut_hash_find_raw(Gnut_Hash *gh, void *data)
{
	int a;
	int ret;
	
	Gnut_List *tmp;
	
	if (!gh) {
	  //		g_debug(0,"gh=%p data=%p\n");
	  //		exit(0);
		g_debug(5,"not found\n");	
		return NULL;
	}

	for (a = 0; a < 256; a++) {
	
	  g_debug(5,"hash returned: %i\n",a);
	
	  for (tmp=gh->list[a];tmp;tmp=tmp->next) {
	    ret=(*gh->compare_func)(data,tmp->data);
	    if (ret==0) return tmp->data;
	  }
	}

	g_debug(5,"not found\n");	
	return NULL;

}

void * gnut_hash_find(Gnut_Hash *gh, void *data)
{
	int a;
	int ret;
	Gnut_List *tmp;
	
	if (!gh) {
		g_debug(0,"gh=%p data=%p\n",gh,data);
		exit(0);
	}
		
	a=(*gh->hash_func)(data);
	
	g_debug(5,"hash returned: %i\n",a);
	
	for (tmp=gh->list[a];tmp;tmp=tmp->next) {
		gnut_mprobe(tmp);
		ret=(*gh->compare_func)(data,tmp->data);
		if (ret==0) return tmp->data;
	}

	g_debug(3,"not found\n");	
	return NULL;
}

uint32 gnut_hash_remove(Gnut_Hash *gh, void *data)
{
	int a;
	
	if (!gh) {
		g_debug(0,"gh=%p data=%p\n",gh,data);
		exit(0);
	}
	
	g_debug(3,"entering\n");
	
	a=(*gh->hash_func)(data);
	
	gh->list[a]=gnut_list_remove(gh->list[a],data);
	
	return 0;
}

uint32 gnut_hash_free(Gnut_Hash *gh)
{
	int i;
	
	g_debug(3,"entering gh=%p\n",gh);
	
	if (!gh) {
		g_debug(0,"gh=%p\n",gh);
		exit(0);
	}
	
	for (i=0;i<256;i++) gh->list[i]=gnut_list_free(gh->list[i]);
	
	xfree(gh);
	return 0;
}
