/* Josh Pieper, (c) 2000 */

/* This file is distributed under the GPL, see file COPYING for details */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#ifndef WIN32
#include <unistd.h>
#endif
#include <fcntl.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <ctype.h>
#include <pthread.h>

#ifdef HAVE_READLINE
	#include <readline/readline.h>
	#include <readline/history.h>
#endif

#ifdef HAVE_REGEX_H
#include <regex.h>
#endif

#include "gnut_lib.h"
#include "gnut_threads.h"
#include "Gnut_List.h"
#include "Gnut_Hash.h"
#include "conf.h"
#include "gnut_connection.h"
#include "gnut.h"
#include "host.h"
#include "share.h"
#include "cache.h"
#include "query.h"
#include "monitor.h"
#include "gnut_transfer.h"
#include "gnut_net.h"
#include "cli_output.h"
#include "cli_input.h"
#include "share_cache.h"
#if GNUT_HTTP_FILE_LIST || GNUT_HTTP_SEARCH
	#include "gnut_http.h"
#endif

int com_info(char *), com_open(char  *), com_find(char *),
    com_update(char *),com_response(char *), com_get(char *),
    com_stop(char *), com_kill(char *), com_hosts(char *),
    com_limit(char *), com_log(char *), com_set(char *), 
    com_quit(char *), com_help(char *), com_version(char *),
    com_clear(char *), com_share(char *), com_scan(char *),
    com_push(char *), com_monitor(char *), com_play(char *),
    com_debug(char *);
    
COMMAND commands[] = {
	{"info", 	com_info, 	"Display various information." },
	{"open", 	com_open, 	"Open a new connection." },
	{"find", 	com_find,	"Search the Gnutella network." },
	{"update",	com_update,	"Update the current host list." },
	{"response",com_response,"Show current query responses." },
	{"get",		com_get,	"Start download(s)." },
	{"push",	com_push,	"Request a push connection."},
	{"stop",	com_stop,	"Stop transfer(s)." },
	{"clear",	com_clear,	"Clear finished transfer(s)." },
	{"kill",	com_kill,	"Terminate connection(s)." },
	{"hosts",	com_hosts,	"Display host catcher." },
	{"limit",	com_limit,	"Impose bandwidth limits on transfers." },
	{"log",		com_log,	"Set logging level." },
	{"share",	com_share,	"Set share paths." },
	{"scan",	com_scan,	"Rescan share paths."},
	{"set",		com_set,	"Change configuration values." },
	{"version",	com_version,"Display version information." },
	{"help",	com_help,	"Display help." },
	{"quit",	com_quit,	"Quit." },
    {"monitor",	com_monitor,	"Monitor search requests."},
    {"play",	com_play,	"Start play_prog on selected queries."},
    {"debug",   com_debug,  "Do whatever I feel like today."},
	{ (char *) NULL, (int (*)(char *)) NULL, (char *) NULL }
};

int get_command_index(char *command)
{
	int i;
	int flag;
	
	for (i=0,flag=-1;commands[i].name!=NULL;i++) {
		if (strncasecmp(command,commands[i].name,strlen(command))==0) {
			if (flag!=-1) return -2;
			flag=i;
		}
	}
	return flag;
}

int parse_command(char *a)
{
	char *command;
	char *arg;
	int cnum;
	int i;
	int len;
	
	if (a[0]=='#') return 0;
	
	len=strlen(a);
	
	for (i=0;i<len && isspace(a[i]);i++);
	if (i==len) {
		command=NULL;
		arg=NULL;
	} else {
		command=&a[i];
		for (;i<len && !isspace(a[i]);i++);
		if (i==len) {
			arg=NULL;
		} else {
			a[i++]=0;
			for (;i<len && isspace(a[i]);i++);
			if (i==len) arg=NULL;
				else arg=&a[i];
		}
	}
	
	if (arg!=NULL) {
		for (i=len;i>0 && isspace(a[i]);i--);
		if (i!=len) a[i+1]=0;
	}
	
	if (command!=NULL) {
		cnum=get_command_index(command);
		if (cnum==-1) {
			printf("no such command: %s\n",command);
			return 0;
		}
		if (cnum==-2) {
			printf("ambiguous command: %s\n",command);
			return 0;
		}
		
		(*commands[cnum].func)(arg);
	}
	return 0;
}

#ifdef HAVE_READLINE
int command_loop()
{
	char *a;
	int ret=0;
	printf("gnut version %s at your service\n\n",VERSION);
	
	while (ret>=0) {
		a=readline("gnut> ");
		if (a==NULL) return 0;
		if (strlen(a)>=1) {
			add_history(a);
			ret=parse_command(a);
		}
		xfree(a);
	}
	
	return 0;
}
#else
int command_loop()
{
	char buf[256];
	char *a;
	printf("gnut version %s at your service\n\n",VERSION);
	
	do {
		printf("gnut> ");
		fflush(stdout);
		
		a=fgets(buf,256,stdin);
		
		if (a!=NULL) {
			buf[strlen(buf)-1]=0;
			if (strlen(buf)>=1) {
				parse_command(buf);
			}
		}
	} while (a!=NULL);
	return 0;
}
#endif

int run_config_file(char *f)
{
	FILE *fp;
	char buf[256];
	
	fp=fopen(f,"r");
	if (fp==NULL) return -1;
	
	while (fgets(buf,sizeof(buf),fp)!=NULL) {
		if (buf[strlen(buf)-1]=='\n') buf[strlen(buf)-1]=0;
		parse_command(buf);
	}
	
	return 0;
}

// parses range, which can have commas and dashes, cand call
// a for each number that results...
int input_parse_range(char *range, int (*a)(int, void *),void *userdata)
{
	int i;
	int r1,r2;
	char *ptr;
	char *ptr2;
	
	if (range==NULL) return 0;
	
	ptr=range;
	while ((*ptr)!=0) {
		g_debug(5,"here\n");
		for (;!isdigit(*ptr) && *ptr!=0;ptr++); // get past any initial whitespace
		if (*ptr==0) return -1;
		r1=strtol(ptr,&ptr2,10);
		g_debug(5,"here\n");
		ptr=ptr2;
		
		// now that we have one number, we need to look for either
		// a control character [,-] or end of string, otherwise it's
		// an error
		
		for (;isspace(*ptr) && *ptr!=0 ;ptr++);
		g_debug(5,"here\n");
		if (*ptr==0) {
			// this was the last number
			(*a)(r1,userdata);
			return 0;
		} else if (*ptr=='-') {
			g_debug(5,"here\n");
			// this is a range, so we need to find the second number....
			// if there is one, then do the whole range
			for (ptr++;!isdigit(*ptr) && *ptr!=0;ptr++);
			if (*ptr==0) return -1; // no second number!
			g_debug(5,"here\n");
			
			r2=strtol(ptr,&ptr2,10);
			ptr=ptr2;
			// k, we have a second number, so lets do the range!!!
			
			for (i=r1;i<=r2;i++) (*a)(i,userdata);
			
			// now we need to see if the string is ended, or if there is a
			// comma yet...
			for (;*ptr!=',' && *ptr!=0;ptr++);
			if (*ptr==0) return 0;
			ptr++;
		} else if (*ptr==',') {
			(*a)(r1,userdata);
			ptr++;
		}
	}		

	return 0;
}

char *info_str_type(int a)
{
	switch(a) {
	case 4:
		return "RCV PUSH";
	case 3:
		return "SND PUSH";
	case 2:
		return "IN";
	case 1:
		return "OUT";
	}
	return NULL;
}

char *info_str_state(int a)
{
	switch(a) {
	case STATE_CONNECTING:
		return "CONN";
	case STATE_NEGOTIATING:
		return "NEGOT";
	case STATE_CONNECTED:
		return "UP";
	case STATE_CLEANABLE:
		return "CLEAN";
	case STATE_DEAD:
		return "DONE";
	case STATE_ERROR:
		return "ERR";
	case STATE_QUEUED:
		return "QUEUED";
	}
	return "Unknown State";
}

int transfer_num=0;
int connection_num=0;
float inb=0;
float outb=0;

int info_transfer_callback(gnut_transfer *gt)
{
	float b,c;
	time_t d;
	char *t1,*t2,*t3,*eta;
	char buf[256];
	char ipstr[32];
	float eta_in_seconds;
	
	transfer_num++;
	gt->index=transfer_num;
	d=time(NULL) - gt->rate_start_time;
	if (d!=0) b=((float) gt->rate_bytes / (float) d);
		else b=0;
	t1=munge_bytes((unsigned long)b);
	if (gt->total!=0) c=100.0f*((float) gt->bytes / (float) gt->total);
		else c=0;

	if (b>0) eta_in_seconds=((gt->total-gt->bytes)/b);
		else eta_in_seconds=0;
	eta=munge_time(eta_in_seconds);
	
	sprintf(ipstr,"%i.%i.%i.%i:%i",gt->ip[0],gt->ip[1],gt->ip[2],gt->ip[3],gt->port);

	if (conf_get_int("munge")) {
		t2=munge_bytes(gt->bytes);
		t3=munge_bytes(gt->total);
	} else {
		t2=xmalloc(10);
		t3=xmalloc(10);
		sprintf(t2,"%i",gt->bytes);
		sprintf(t3,"%i",gt->total);
	}

	sprintf(buf,"%i)%-19s "
	    "%5.1f%%  %9s/%-9s  %6s/sec    ETA: %s\n",
		gt->index,ipstr,
		c,t2,t3,t1,eta);
	xfree(t1);
	xfree(t2);
	xfree(t3);
	xfree(eta);
	
	if(cli_output_line(1,buf)) return -1;
	
	sprintf(buf,"  TID: %-6li  %3s  State: %-4s Name: ", 
#if defined(hpux) && defined(PTHREADS_DRAFT4)
		((HPUX_hack)gt->tid).longpart,
#else
		(long) gt->tid,
#endif
		info_str_type(gt->type),info_str_state(gt->state));
		
	if (cli_output_line(1,buf)) return -1;	
	tputs_x(output_bold_on);
	t1=(gt->fname) ? strdup(gt->fname) : strdup("UNKNOWN");
	t2=strstr(t1,".gnut");
	if (t2) *t2=0;
	sprintf(buf,"%s",t1);
	xfree(t1);
	if (cli_output_line(1,buf)) return -1;
	tputs_x(output_special_off);
	if (cli_output_line(1,"\n")) return -1;

	if (gt->state==STATE_CONNECTED) {
		if ((gt->type&1)==0) inb+=b;
			else outb+=b;
	}
	return 0;
}

int info_transfer()
{
	char buf[256];
	char *t1,*t2;

	if (gnut_transfer_num()==0) return 0;
	if (cli_output_line(3,"TRANSFER_STATS:\n")) return 0;
	cli_output_line(2,"---------------\n");
	transfer_num=0;
	inb=0;
	outb=0;
	gnut_transfer_enumerate(info_transfer_callback);
	t1=munge_bytes(outb);
	t2=munge_bytes(inb);
	sprintf(buf,"TOTALS: Rate: %5s:%-5s /sec\n\n",t1,t2);
	xfree(t1);
	xfree(t2);
	if (cli_output_line(0,buf)) return -1;
	return 0;
}

int info_connection_callback(gnut_connection *gc)
{
	float s,r;
	time_t i;
	int n,n2;
	char *t1,*t2;
	char buf[4096];
	char ipstr[32];
	
	connection_num++;
	gc->index=connection_num;
	
	sprintf(ipstr,"%i.%i.%i.%i:%i",
		gc->ip[0],gc->ip[1],gc->ip[2],gc->ip[3],gc->port);
	
	if (conf_get_int("munge")) {
		t1=munge_bytes(gc->sent_bytes);
		t2=munge_bytes(gc->received_bytes);
	} else {
		t1=xmalloc(20);
		sprintf(t1,"%7i",gc->sent_bytes);
		t2=xmalloc(20);
		sprintf(t2,"%-7i",gc->received_bytes);
	}
	sprintf(buf,"%i)%-20s"
	    "  Packs: %6i:%-6i %5i:%-3i \tBytes: %7s:%-7s\n",
		connection_num,ipstr,
		gc->sent_packets,
		gc->received_packets,gc->dropped_packets,gc->bad_packets,
		t1,t2);
		
	xfree(t1);
	xfree(t2);
	if (cli_output_line(1,buf)) return -1;
	
	
	i=time(NULL) - gc->rate_start_time;
	if (i!=0) {
		s=(float) gc->rate_sent_bytes / (float) i;
		t1=munge_bytes((unsigned long)s);

		r=(float) gc->rate_received_bytes / (float) i;
		t2=munge_bytes((unsigned long)r);
	} else {
		t1=xmalloc(10);
		t2=xmalloc(10);
		strcpy(t1,"0");
		strcpy(t2,"0");
		s=r=0;
	}
	if (gc->state==STATE_CONNECTED) {
		inb+=r;
		outb+=s;
	}
	sprintf(buf,"   TID: %-6li  Type: %-3s  State: %5s  "
		"Rate: %6s:%-6s /sec\n",
#if defined(hpux) && defined(PTHREADS_DRAFT4)
		((HPUX_hack) gc->tid).longpart,
#else
		(long) gc->tid,
#endif
		info_str_type(gc->type),info_str_state(gc->state),t1,t2);

	xfree(t1);
	xfree(t2);
	if (cli_output_line(0,buf)) return -1;

	
#ifdef MOREPACKETSTATS
	if (conf_get_int("packet_stats")) {
		for (n = 0; n < 256; n++) {
			if (gc->packet_count[n]) {
				sprintf(buf,"\tIn:\tPacket type %3d: %5d   avg %5d  max %5d\n",
					   n, gc->packet_count[n], gc->packet_byte_count[n]/gc->packet_count[n],
					   gc->max_packet_size[n]);
				if (cli_output_line(0,buf)) return -1;
			}
		}

		if (gc->packet_count[0]) {
			n2=sprintf(buf,"\tIn:\tTTL ");
			for (n = 0; n < 256; n++) {
				if (gc->ttl_count[n]) {
					n2+=sprintf(&buf[n2],"%d(%d) ",
						   n, gc->ttl_count[n]);
				}
			}
			sprintf(&buf[n2],"\n");
			if (cli_output_line(0,buf)) return -1;
		}

		for (n = 0; n < 256; n++) {
			if (gc->out_packet_count[n]) {
				sprintf(buf,"\tOut:\tPacket type %3d: %5d   avg %5d  max %5d\n",
					   n, gc->out_packet_count[n], gc->out_packet_byte_count[n]/gc->out_packet_count[n],
					   gc->out_max_packet_size[n]);
				if (cli_output_line(0,buf)) return -1;
			}
		}

		if (gc->out_packet_count[0]) { 
			n2=sprintf(buf,"\tOut:\tTTL ");
			for (n = 0; n < 256; n++) {
				if (gc->out_ttl_count[n]) {
					n2+=sprintf(&buf[n2],"%d(%d) ",
						   n, gc->out_ttl_count[n]);
				}
			}
			sprintf(&buf[n2],"\n");
			if (cli_output_line(0,buf)) return -1;
		}
	}
#endif
		
	return 0;
}

int info_connection()
{
	char buf[256];
	char *t1,*t2;
	
	if (gnut_connection_num()==0) return 0;
	
	if (cli_output_line(3,"CONNECTION STATS:\n")) return -1;
	cli_output_line(2,"-----------------\n");
	connection_num=0;
	inb=0;
	outb=0;
	gnut_connection_enumerate(info_connection_callback);
	
	t1=munge_bytes((unsigned long) outb);
	t2=munge_bytes((unsigned long) inb);
	sprintf(buf,"TOTALS: Rate: %5s:%-5s /sec\n\n",t1,t2);
	xfree(t1);
	xfree(t2);
	if (cli_output_line(0,buf)) return -1;
	return 0;
}

int info_shares()
{
	int num,size;
	char *t1;
	char buf[256];
	
	share_totals(&num,&size);
	t1=munge_bytes((uint64)size*(uint64)1024);
	sprintf(buf,"SHARE STATS: Num Shared: %-5i Size Shared: %-7s\n",num,t1);
	xfree(t1);
	if (cli_output_line(0,buf)) return -1;
	return 0;
}

int info_net()
{
	int messages, sent;
	char *t1,*t2;
	uint64 rcv_bytes,sent_bytes;
	char buf[256];
	
	gnut_connection_net_totals(&messages,&sent,&rcv_bytes,&sent_bytes);
	t1=munge_bytes(messages);
	t2=munge_bytes(sent);
	sprintf(buf,"NET STATS:   Msg Received: %-7s  Msg Sent: %-7s\n",
		t1,t2);
	xfree(t1);
	xfree(t2);
	if (cli_output_line(1,buf)) return -1;
	
	t1=munge_bytes(rcv_bytes);
	t2=munge_bytes(sent_bytes);
	sprintf(buf,"             Bytes Rcvd: %-7s       Bytes Sent: %-7s\n",
		 t1,t2);
	xfree(t1);
	xfree(t2);
	if (cli_output_line(0,buf)) return -1;
	return 0;
}

int info_query()
{
	int queries,responses;
	char buf[256];
	
	gnut_connection_query_totals(&queries,&responses);
	sprintf(buf,"QUERY STATS: Queries: %-6i   Responses Sent: %-6i\n",
		queries,responses);
	if (cli_output_line(0,buf)) return -1;
	return 0;
}

int info_hosts()
{
	uint num;
	uint64 size,files;
	char buf[256];
	char *t1,*t2;
	host_totals(&num,&files,&size);
	
	t1=munge_bytes(files);
	t2=munge_bytes(size*(uint64)1024000);
	sprintf(buf,"HOST STATS:  Hosts: %-6i Files: %-7s Size: %-7s\n",
		num,t1,t2);
	xfree(t1);
	xfree(t2);
	if (cli_output_line(0,buf)) return -1;
	return 0;
}	

int info_all()
{
	cli_output_reset();
	if (info_hosts()) return 0;
	if (info_net()) return 0;
	if (info_query()) return 0;
	if (info_shares()) return 0;
	if (info_connection()) return 0;
	if (info_transfer()) return 0;
	return 0;
}

int com_info(char *arg)
{
	if (arg==NULL) {
		info_all();
	} else {
		switch(tolower(arg[0])) {
		case 'h':
			cli_output_reset();
			info_hosts();
			break;
		case 'q':
			cli_output_reset();
			info_query();
			break;
		case 'n':
			cli_output_reset();
			info_net();
			break;
		case 's':
			cli_output_reset();
			info_shares();
			break;
		case 'c':
			cli_output_reset();
			info_connection();
			break;
		case 't':
			cli_output_reset();
			info_transfer();
			break;
		}
	}
	return 0;
}

int com_open(char *arg)
{
	if (arg!=NULL) {
		gnut_outgoing_new(arg);
	}
	return 0;
}

void show_results_count()
{
	static int a;
	int b;
	
	b=query_num();
	if (b==0) a=0;
	if (b!=0 && a==0) {
		if (conf_get_int("beep_on_first")) {
			putchar(7);
		}
	}
	
	printf("%i results received.",query_num());
	fflush(stdout);
	printf("\r");
	a=b;
}	

int com_find(char *arg)
{
	gnutella_packet *gpa;
	
	if (arg!=NULL) {
//		gpa=gp_request_make(conf_get_mac(),arg,conf_get_ttl());
		gpa=gp_request_make(conf_get_str("mac"),arg,conf_get_int("ttl"));
//		conf_set_query_guid(gpa->gh.guid);
		conf_set_str_len("query_guid",gpa->gh.guid,16);
		query_clear();
		send_to_all(gpa);
		xfree(gpa->data);
		xfree(gpa);
		printf("Searching the gnutella network for: %s \nPress any key to continue\n", arg);

		wait_key_callback(show_results_count);

		printf("\n");
		if (conf_get_int("auto_response")) {
			com_response(NULL);
		}
	}
	return 0;
}

int com_update(char *arg)
{
	gnutella_packet *gpa;
	
//	gpa=gp_ping_make(conf_get_mac(),conf_get_ttl());
	gpa=gp_ping_make(conf_get_str("mac"),conf_get_int("ttl"));
	if (conf_get_int("update_clear")) host_clear();
	
//	conf_set_update_guid(gpa->gh.guid);
	conf_set_str_len("update_guid",gpa->gh.guid,16);
		
	send_to_all(gpa);
	
	if (gpa->data) xfree(gpa->data);
	xfree(gpa);
	return 0;
}

int query_response_compare(const void *a, const void *b)
{
	query_response **c;
	query_response *qr1,*qr2;
	char *so;
	char *e1,*e2;
	int flag=0;
	
	c=(query_response **) a;
	qr1=*c;
	c=(query_response **) b;
	qr2=*c;
	
	so=conf_get_str("sort_order");
	
	while (*so) {
		switch(*so) {
		case 's':
			if (qr1->size < qr2->size) flag=-1;
			if (qr1->size > qr2->size) flag=1;
			break;
		case 'p':
			if (qr1->speed < qr2->speed) flag=-1;
			if (qr1->speed > qr2->speed) flag=1;
			break;
		case 'e':
			e1=strrchr(qr1->name,'.');
			if (!e1) break;
			e2=strrchr(qr2->name,'.');
			if (!e2) break;
			flag=strcasecmp(e1,e2);
			break;
		case 'n':
			flag=strcasecmp(qr1->name,qr2->name);
			break;
		}
		if (flag) break;
		if (!(*(++so))) break;
		so++;
	}
	if (*so) {
		so++;
		if ((*so)=='-') flag=flag*-1;
	}
	
	return flag;
}

// I've got some new logic for the reponses.
// First a count of all eligible responses is made, this is done by 
// running the current expression through the entire list, incrementing
// 1 each time.  This is limited by "max_responses" in the config.
// After a count is made, an array of pointers to query_response strucs
// is allocated, and the results are copied into it.  Then qsort is used
// to sort this array.  Finally it is displayed, and the array is free'ed
// once displaying has ended.

int com_response(char *arg)
{
	Gnut_List *tmp;
	query_response *qr;
	uint16 port;
	int index=1;
	int cre=0;		/* use re */
	int ig_case=1;		/* ignore case in re */
	int inv_re=0;		/* invert RE matches */
	char *re = NULL;
	char *t1;
	char buf[256];
	char ipstr[32];
	query_response **disp_qr_list=NULL;
	int result_count=0;

/* regcomp is much more full featured than re_comp, so prefer it if both
** are available.
 */
#ifdef HAVE_REGCOMP
	regex_t creg;
	int re_opt;
#else
  #ifdef HAVE_RE_COMP
	char * cre_err = NULL;
  #endif
#endif
	cli_output_reset();


	if (arg && (*arg)) {
		if (('/' == arg[0]) && arg[1]) {
			int in,out;
			re = xmalloc(strlen(arg));

			for(in=1,out=0; arg[in] && (arg[in] != '/'); in++) {
				if(('\\' == arg[in]) && arg[in+1]) {
					switch (arg[in+1]) {
						case '\\':
							re[out] = '\\';
							out ++;
							in  ++;
							break;
						case '/':
							re[out] = '/';
							out ++;
							in  ++;
							break;
					}
				} else {
					re[out] = arg[in];
					out++;
				}
			}

			re[out] = '\0';

			if (arg[in] && arg[in+1]) {
				in++;
				while(arg[in]) {
					switch (arg[in]) {
						case 'i':
							break;
						case 'I':
							ig_case=0;
							break;
						case 'v':
							inv_re=1;
							break;
					}
					in++;
				}
			}

		} else {
			re = arg;
		}

#ifdef HAVE_REGCOMP
		if (ig_case) {
			re_opt = REG_EXTENDED|REG_ICASE|REG_NOSUB;
		} else {
			re_opt = REG_EXTENDED|REG_NOSUB;
		}
		cre = !regcomp(&creg,re,re_opt);
#else
# ifdef HAVE_RECOMP
		if (ig_case) {
			fprintf(stderr, "re_comp does cannot ignore case\n");
		}
		cre_err=(char *) re_comp(re);
		cre=(cre_err==NULL);
# else
		cre=0;
# endif
#endif
		if (cre) {
			sprintf(buf,"CURRENT RESPONSES for %s\n", arg);
		} else {
			sprintf(buf,"CURRENT RESPONSES (%s not valid RE)\n", arg);
		}
	} else {
		sprintf(buf,"CURRENT RESPONSES\n");
	}
	cli_output_line(3,buf);
	
	cli_output_line(2,"-----------------\n");
	
	query_lock();

	// here we count the number of responses that meet the criteria
	for (tmp=query_retrieve();tmp;tmp=tmp->next) {
		int result=1;
		qr=tmp->data;
		qr->index=-1;
//		qr->index=index++;
		if (cre) {
#ifdef HAVE_REGCOMP
			result = !regexec(&creg,qr->name,0,NULL,0);
#else
#  ifdef HAVE_RE_COMP
			result = re_exec(qr->name);
#  else
			result = is_match_tok(qr->name,re);
#  endif
#endif
			if (inv_re) {
				result = !result;
			}
		}

		if (result) result_count++;
	}		

	if (result_count>conf_get_int("max_responses")) result_count=conf_get_int("max_responses");

	// now that we have the list counted, we can allocate the array
	disp_qr_list=(query_response **) xmalloc(result_count*sizeof(query_response *));
	
	// now we have to go back through the list and fill in the array with
	// pointers....
	index=0;
	for (tmp=query_retrieve();tmp && index<result_count;tmp=tmp->next) {
		int result=1;
		qr=tmp->data;
		if (cre) {
#ifdef HAVE_REGCOMP
			result = !regexec(&creg,qr->name,0,NULL,0);
#else
#  ifdef HAVE_RE_COMP
			result = re_exec(qr->name);
#  else
			result = is_match_tok(qr->name,re);
#  endif
#endif
			if (inv_re) {
				result = !result;
			}
		}

		if (result) disp_qr_list[index++]=qr;
	}		

	// now that we have the list of reponses in the array, we can unlock
	// the query list
	
	query_unlock();	

	// now we need to sort the list
	
	qsort(disp_qr_list,result_count,sizeof(query_response *),
		query_response_compare);

	for (index=0;index<result_count;index++) {
		qr=disp_qr_list[index];
		qr->index=index+1;
		sprintf(buf,"%i)",qr->index);
		if (cli_output_line(1,buf)) break;
		
		tputs_x(output_bold_on);
		sprintf(buf,"%s",qr->name);
		if (cli_output_line(1,buf)) break;
		tputs_x(output_special_off);
		if (cli_output_line(1,"\n")) break;
			
		memcpy(&port,qr->port,2);
		port=GUINT16_FROM_LE(port);
		sprintf(ipstr,"%i.%i.%i.%i:%i",qr->ip[0],qr->ip[1],qr->ip[2],qr->ip[3],port);
		
		if (conf_get_int("munge")) {
			t1=munge_bytes(qr->size);
		} else {
			t1=xmalloc(15);
			sprintf(t1,"%i",qr->size);
		}
		sprintf(buf," %-20s  size:%-9s  ref:%4i  speed:%5i\n",
			ipstr,
			t1,qr->ref,qr->speed);
		xfree(t1);
		if (cli_output_line(0,buf)) break;;
	}

	xfree(disp_qr_list);
	
#ifdef HAVE_REGCOMP
	if (arg && (*arg)) regfree(&creg);
#endif
	if (arg && (*arg) && (re != arg)) xfree(re);
	return 0;
}

int com_push_callback(int a, void *b)
{
	query_response *qr;
	gnutella_packet *gpa;
	
	qr=query_index(a);
	
	if (qr==NULL) return 0;
	
	gpa=gp_push_make(conf_get_str("mac"),conf_get_int("ttl"),
		qr->guid,qr->ref,net_local_ip(),conf_get_int("local_port"));

	query_push_add(qr);	
		
	send_to_all(gpa);
	
	xfree(gpa->data);
	xfree(gpa);
	
	return 0;
}


int com_push(char *a)
{
	input_parse_range(a,com_push_callback,NULL);
	return 0;
}

int com_get_callback(int a, void *b)
{
	query_response *qr;
	
	qr=query_index(a);
	
	if (qr==NULL) {
		printf("No such ref num: %i\n",a);
		return 0;
	}
	
	gnut_transfer_start(qr);
	
	return 0;
}

int com_get(char *arg)
{
	input_parse_range(arg,com_get_callback,NULL);
	return 0;
}

int com_stop_index=0;

int com_stop_callback2(gnut_transfer *gt)
{
	if ((gt->index==com_stop_index || com_stop_index==-1) && gt->state==STATE_CONNECTED) {
		printf("Stopping transfer: %i\n",gt->index);
		gt->state=STATE_DEAD;
	}
	return 0;
}

int com_stop_callback(int a, void *b)
{
	com_stop_index=a;
	gnut_transfer_enumerate(com_stop_callback2);
	return 0;
}
	

int com_stop(char *arg)
{
	if (arg!=NULL) {
		input_parse_range(arg,com_stop_callback,NULL);
	} else {
		com_stop_index=-1;
		gnut_transfer_enumerate(com_stop_callback2);
	}
	return 0;
}

int com_clear_index=0;

int com_clear_callback2(gnut_transfer *gt)
{
	if ((gt->index==com_clear_index || com_clear_index==-1) && (gt->state==STATE_DEAD || gt->state==STATE_ERROR)) {
		printf("Clearing transfer: %i\n",gt->index);
		gt->state=STATE_CLEANABLE;
	}
	return 0;
}

int com_clear_callback(int a, void *b)
{
	com_clear_index=a;
	gnut_transfer_enumerate(com_clear_callback2);
	return 0;
}

int com_clear(char *arg)
{
	if (arg!=NULL) {
		input_parse_range(arg,com_clear_callback,NULL);
	} else {
		com_clear_index=-1;
		gnut_transfer_enumerate(com_clear_callback2);
	}
	return 0;
}

int com_kill_index=0;

int com_kill_callback2(gnut_connection *gc)
{
	if (gc->index==com_kill_index || com_kill_index==-1) {
		printf("Killing %i\n",gc->index);
		gc->state=STATE_DEAD;
	}
	return 0;
}

int com_kill_callback(int a, void *b)
{
	com_kill_index=a;
	gnut_connection_enumerate(com_kill_callback2);
	return 0;
}

int com_kill(char *arg)
{
	if (arg!=NULL) {
		input_parse_range(arg,com_kill_callback,NULL);
	} else {
		com_kill_index=-1;
		gnut_connection_enumerate(com_kill_callback2);
	}
	return 0;
}

int host_disp(void *a, void *b)
{
	host_entry *he;
	uint16 port;
	char ipstr[32];
	char buf[256];
	
	he=a;
	memcpy(&port,he->port,2);
	port=GUINT16_FROM_LE(port);
	
	sprintf(ipstr,"%i.%i.%i.%i:%i",he->ip[0],he->ip[1],he->ip[2],he->ip[3],port);
	sprintf(buf," %20s  [%5d, %8d]\n",ipstr,he->files,he->bytes);
	
	return cli_output_line(0,buf);
}
	

int com_hosts(char *arg)
{
	Gnut_Hash *gl;
	
	if (arg) {
		return host_restore(arg);
	}
	
	host_lock();
	// locking for a long time, but oh well, it's the user's fault..! ;)
	
	gl=host_retrieve();
	gnut_hash_foreach(gl,host_disp,NULL);
	
	host_unlock();
	return 0;
}

int com_limit_amount=0;
int com_limit_index=0;

int com_limit_callback2(gnut_transfer *gt)
{
	if (gt->index==com_limit_index || com_limit_index==-1) {
		printf("Limiting %i to %i bytes/s %i\n",gt->index,com_limit_amount,gt->bytes);
		gt->rate_limit=com_limit_amount;
		gt->rate_start_time=time(NULL);
		gt->rate_bytes=0;
	}
	return 0;
}

int com_limit_callback(int a, void *b)
{
	com_limit_index=a;
	gnut_transfer_enumerate(com_limit_callback2);
	return 0;
}

int com_limit(char *arg)
{
	char *ptr,*ptr2;
	if (!arg) return 0;
	
	ptr=arg;
	for (;!isdigit(*ptr) && *ptr!=0;ptr++);
	if (ptr==0) return -1;
	
	com_limit_amount=strtol(ptr,&ptr2,10);
	for (ptr=ptr2;!isdigit(*ptr) && *ptr!=0;ptr++);
	if (*ptr!=0) {
		input_parse_range(ptr2,com_limit_callback,NULL);
	} else {
		com_limit_index=-1;
		gnut_transfer_enumerate(com_limit_callback2);
	}
	
	
	return 0;
}

int com_log(char *arg)
{
	if (arg==NULL) {
		printf("Current log level is: %i\n",gnut_lib_debug);
	} else {
		printf("New log level is: %i\n",gnut_lib_debug=atoi(arg));
	}
	return 0;
}

int com_share(char *arg)
{
	
//	conf_set_share_paths(b);
	if (arg) conf_set_str("share_paths",arg);
		else conf_set_str("share_paths","");
	return 0;
}

int com_scan(char *arg)
{
	char *a,*b,*c;
	int num, size;
	char *path;

#if GNUT_HTTP_FILE_LIST || GNUT_HTTP_SEARCH
	gnut_http_template_scan();	
#endif

	g_debug(5,"calling share_clear\n");
	share_clear();
	g_debug(5,"done calling share_clear\n");

	if ((b=strdup(conf_get_str("share_paths")))!=NULL) {
		g_debug(5,"working through paths\n");
#ifdef HAVE_STRTOK_R
		a=strtok_r(b,path_separator,&c);
#else
		a=strtok(b,path_separator);
#endif
		while (a!=NULL) {
			path=expand_path(a);
			g_debug(1,"scanning directory: %s\n",path);
			share_scan_dir(path);
#ifdef HAVE_STRTOK_R
			a=strtok_r(NULL,path_separator,&c);
#else
			a=strtok(NULL,path_separator);
#endif
			xfree(path);
		}
	
		xfree(b);
	}
	
	share_totals(&num,&size);
	
	cache_snapshot();
	b=conf_get_str("cache_path");
	if ((b) && (strlen(b)>1)) {
		c=expand_path(b);
		g_debug(2,"scanning cache: %s\n",c);
		share_scan_dir(c);
		xfree(c);
	}
	
	printf("Scanned %i files and %i Kbytes.\n",num,size);	
	return 0;
}

int cli_disp_conf(void *a, void *b)
{
	conf_key_pair *ckp;
	char buf[256];
	
	ckp=a;
	if (!strcmp(ckp->key,"update_guid")) return 0;
	if (!strcmp(ckp->key,"query_guid")) return 0;
	if (!strcmp(ckp->key,"mac")) return 0;
	if (!strcmp(ckp->key,"guid")) return 0;

	sprintf(buf,"%30s = %s\n",ckp->key,ckp->val);
	return cli_output_line(0,buf);
}

int com_set(char *arg)
{
	char *arg2;
	char *b;
	int i;
	
	
	if (arg==NULL) {
		// display a list of all config values
		cli_output_reset();
		gnut_hash_foreach(conf_hash,cli_disp_conf,NULL);
		return 0;
	}

	for (arg2=arg;*arg2 && !isspace(*arg2);arg2++);
	if (*arg2==0) arg2=NULL;
		else {
			*arg2=0;
			arg2++;
			for (;(*arg2) && isspace(*arg2);arg2++);
			if ((*arg2)==0) arg2=NULL;
		}

	if (arg2==NULL) {
		// we'll display the current value...
		b=conf_get_str(arg);
		if (b==NULL) {
			printf("No such configuration option: %s\n",arg);
		} else {
			printf("Config item: %s = %s\n",arg,b);
		}
	} else {
		i=conf_set_str(arg,arg2);
		if (i<0) {
			printf("No such configuration option: %s\n",arg);
		} else {
			printf("Config item: %s = %s\n",arg,arg2);
		}
	}
	return 0;
}

int com_quit(char *arg)
{
	char *d;
#ifdef DMALLOC
	dmalloc_shutdown();
#endif

	g_debug(3,"calling expand_path\n");
	d=expand_path("~/.gnut_hosts");
	g_debug(3,"calling host_save\n");
	host_save(d);
	g_debug(3,"free'ing mem\n");
	xfree(d);
	exit(0);
	return 0;
}

int com_help(char *arg)
{
	int cnum;
	printf("\n");
	
	if (arg!=NULL) {
		cnum=get_command_index(arg);
		if (cnum>=0)
			printf("  %s: %s\n",commands[cnum].name,commands[cnum].doc);
	} else {
		for (cnum=0;commands[cnum].name!=NULL;cnum++) {
			printf("  %s: %s\n",commands[cnum].name,commands[cnum].doc);
		}
	}
	printf("\n");
	return 0;
}

int com_version(char *arg)
{
	printf("gnut version %s (c) 2000 Josh Pieper\n",VERSION);
	return 0;
}

int monitor_search_requests = 0;

void monitor_callback()
{
	monitored_search *ms;
	while((ms = monitor_search_get())) {
		printf("Search: %s%s\n",  ms->search,
		ms->matched?" (matched)":"");
		xfree(ms->search);
		xfree(ms);
		fflush(stdout);
	}
}
	

int com_monitor(char *arg)
{
	monitored_search *ms;

	// tell the packet-handling thread to watch the searches
	monitor_search_requests = 1;

	printf("Monitoring search requests.\nPress any key to continue\n");

	wait_key_callback(monitor_callback);

	printf("\n");

	// tell the other threads to stop adding the searches to the list
	monitor_search_requests = 0;

	// delete the rest of the list
	while((ms = monitor_search_get())) {
		xfree(ms->search);
        xfree(ms);
	}

	return 0;
}


static char **com_play_scratch=NULL;
static int com_play_scratch_size=0,com_play_scratch_index;

void com_play_append_string(char *str)
{
	/* make sure we have enough room */
	if ((com_play_scratch_index +1) >= com_play_scratch_size) {
		com_play_scratch_size *= 1.5;
		com_play_scratch=realloc(com_play_scratch,com_play_scratch_size);
	}
	
	/* append a copy of this string, make sure the array is terminated
	   (in case of an early bailout in com_play) */
	com_play_scratch[com_play_scratch_index++] = strdup(str);
	com_play_scratch[com_play_scratch_index] = NULL;
}

int com_play_callback(int a, void *arg)
{
	char *tmp;
	query_response *qr;
	uint16 port;
	int len;
	
	qr=query_index(a);
	if (qr==NULL) {
		printf("No such ref num: %i\n",a);
		return 0;
	}
	
	memcpy(&port,qr->port,2);
	port=GUINT16_FROM_LE(port);
	
	len=strlen(qr->name)+60;
	tmp=xmalloc(len);
	sprintf(tmp,"http://%i.%i.%i.%i:%i/get/%i/%s",
		qr->ip[0],qr->ip[1],qr->ip[2],qr->ip[3],
		port,qr->ref,qr->name);
	com_play_append_string(tmp);
	xfree(tmp);
	return 0;
}

int com_play(char *arg)
{
#ifdef WIN32
	return 0;
#else
	int i;
	char *temp, *a, *b;

	
	/* allocate the play command argument storage if necessary */
	if (!com_play_scratch) {
		com_play_scratch_size=100;
		com_play_scratch= (char **) malloc(com_play_scratch_size * sizeof(char *));
	} else {
	
		/* clear any previous commands */
		i=0;
		while (com_play_scratch[i]) {
			xfree(com_play_scratch[i++]);
		}
	}

	com_play_scratch[0]=NULL;

	com_play_scratch_index=0;
	
	temp=strdup(conf_get_str("play_prog"));
#ifdef HAVE_STRTOK_R
	a=strtok_r(temp," \t",&b);
#else
	a=strtok(temp," \t");
#endif
	if (!a) {
		printf("Empty play_prog configured!\n");
		return 0;
	}
	
	while (a) {
		com_play_append_string(a);
#ifdef HAVE_STRTOK_R
		a=strtok_r(NULL," \t",&b);
#else
		a=strtok(NULL," \t");
#endif
	}
	
	xfree(temp);
	
	/* now parse the arguments to "play" parsing out ranges for
	   any argument starting with a digit */
	   
	/* Ray, that strdup wasn't necessary. ;) */
	temp=arg;
	
#ifdef HAVE_STRTOK_R
	a=strtok_r(temp," \t",&b);
#else
	a=strtok(temp," \t");
#endif
	
	while (a) {
		if (isdigit(a[0])) {
			input_parse_range(a,com_play_callback,NULL);
		} else {
			com_play_append_string(a);
		}
#ifdef HAVE_STRTOK_R
		a=strtok_r(NULL," \t",&b);
#else
		a=strtok(NULL," \t");
#endif
	}
	printf("The final command was:");
	i=0;
	while (com_play_scratch[i]) {
		printf(" %s", com_play_scratch[i++]);
	}
	printf("\n");
	
	if (!fork()) {
		execvp(com_play_scratch[0],com_play_scratch);
		printf("Failure executing %s\n",com_play_scratch[0]);
		exit(0);
	}
	
	return 0;
#endif
}
 
int com_debug(char *a)
{
	Gnut_List *gl;
	Gnut_List *tmp;
	share_item *si;
	
	gl=share_cache_run_search("mpg",64);
	
	for (tmp=gl;tmp;tmp=tmp->next) {
		si=tmp->data;
		printf("ref: %i  name: %s\n",si->ref,si->fpath);
	}
	
	share_clear_list(gl);
	return 0;
}
	