/* Josh Pieper, (c) 2000 */

/* This file is distributed under the GPL, see file COPYING for details */

#include <stdio.h>
#include <stdlib.h>
#ifndef WIN32
#include <unistd.h>
#endif

#ifdef HAVE_READLINE
#include "cli_readline.h"
#endif

#include <malloc.h>

#ifdef MCHECK
#include <mcheck.h>
#endif

#ifdef HAVE_GETOPT_H
#include <getopt.h>
#endif

#include "gnut_lib.h"
#include "gnut.h"
#include "conf.h"
#include "gnut_net.h"
#include "cli_input.h"
#include "cli_output.h"

int usage()
{
	printf("Usage gnut [OPTION] ... [HOST] ...\n");
	printf("Client for the Gnutella distributed file sharing network.\n\n");
	printf("  -l [arg]          sets the initial logging level to arg\n");
	printf("  -c [file]         specify a script to run instead of ~/.gnutrc\n");
	printf("  -i [ip/interface] specifies the interface to bind to\n");
	printf("  -p [port]         port to listen on\n");
	printf("  -d                daemon mode, never ask for input\n");
	printf("  -x                exit after having run any scripts\n");
	printf("  -v                output version information and exit\n");
	printf("  -h                display this help and exit\n\n");
	printf("Report bugs to josh@nemonet.com\n");
	return 0;
}

#ifdef MCHECK
void mcheck_callback(enum mcheck_status a)
{
	printf("mcheck called!  a=%i  PID=%i\n",a,getpid());
	while (1) sleep(1);
}
#endif	

int main(int argc, char *argv[])
{
	int daemon=0;
	int ex=0;
	int c;
	char *d;
	char *config_file=NULL;
	char *iface=NULL;
        int option_index = 0;
	char *gnut_short_options = "xdhvc:l:p:i:";


#ifdef HAVE_GETOPT_H
	static struct option gnut_long_options[] =
	{
		{"exit", 0, 0, 'x'},
		{"daemon", 0, 0, 'd'},
		{"help", 0, 0, 'h'},
		{"version", 0, 0, 'v'},
		{"config", 1, 0, 'c'},
		{"loglevel", 1, 0, 'l'},
		{"port", 1, 0, 'p'},
		{"interface", 1, 0, 'i'},
		{"ip", 1, 0, 'i'},
		{0, 0, 0, 0}
	};
#endif

#ifdef MCHECK
	if (mcheck(mcheck_callback)!=0) {
		printf("couldn't do mcheck!\n");
		exit(0);
	}
#endif
	
	conf_init();
	cli_output_init();

#ifdef HAVE_READLINE
	initialize_readline();
#endif
	
#ifdef HAVE_GETOPT_H
	while ((c=getopt_long(argc,argv, gnut_short_options, gnut_long_options,&option_index))!=EOF) {
#else
	while ((c=getopt(argc,argv,gnut_short_options))!=EOF) {
#endif
		switch(c) {
		case 'p':
			conf_set_int("local_port",atoi(optarg));
			break;
		case 'i':
			iface=optarg;
			break;
		case 'x':
			ex=1;
			break;
		case 'd':
			daemon=1;
			break;
		case 'l':
			gnut_lib_debug=atoi(optarg);
			break;
		case 'v':
			printf("gnut version %s at your service\n\n",VERSION);
			return 0;
		case 'c':
			config_file=optarg;
			break;
		case 'h':
			usage();
			return 0;
		}
	}

	gnut_init();
	net_init(iface);
	gnut_start_main_loop();

	if (argv[optind]!=NULL) {
		gnut_outgoing_new(argv[optind]);
	}

	if (config_file!=NULL) {
		d=expand_path(config_file);
		run_config_file(d);
		xfree(d);
	} else {
		d=expand_path("~/.gnutrc");
		run_config_file(d);
		xfree(d);
	}
	
	if (ex==1) exit(0);
	
	if (daemon==0) {		
		command_loop();
	} else {
		printf("gnut %s Entering no output mode, have fun!\n",VERSION);
	}
	while(1) sleep(1);
	
	return 0;
}

	