/* Josh Pieper, (c) 2000 */

/* This file is distributed under the GPL, see file COPYING for details */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#ifndef WIN32
#include <unistd.h>
#include <netinet/in.h>
#include <sys/time.h>
#include <sys/wait.h>
#include <sys/socket.h>
#endif
#include <signal.h>
#include <time.h>
#include <sys/types.h>
#include <pthread.h>

#include "gnut.h"
#include "gnut_threads.h"
#include "gnut_net.h"
#include "conf.h"
#include "host.h"
#include "gnut_connection.h"
#include "gnut_transfer.h"
#include "query.h"
#include "share.h"
#include "monitor.h"
#include "route.h"
#include "share_cache.h"

int global_rate=0;
int new_global_rate;

#ifndef WIN32
pthread_t segv_tid;

int segv_callback2(gnut_transfer *gt)
{
	if (pthread_equal( gt->tid, segv_tid)) {
		printf("The following transfer was the culprit.\n");
		printf("%i.%i.%i.%i:%i ",gt->ip[0],gt->ip[1],gt->ip[2],gt->ip[3],gt->port);
		printf("fsock: %i  fname: %s\n", gt->fsock, gt->fname);
		printf("type: %i  bytes: %i  total: %i\n",gt->type,gt->bytes,gt->total);
		printf("rateb: %i  time: %i  rate_lim: %i\n",gt->rate_bytes, (int) gt->rate_start_time, gt->rate_limit);
		printf("state: %i  index: %i\n",gt->state,gt->index);
	}
	
	return 0;
}

int segv_callback(gnut_connection *gc)
{
	if (pthread_equal(gc->tid, segv_tid) ) {
		// we found it!
		printf("The following connection was the culprit.\n");
		printf("%i.%i.%i.%i:%i ",gc->ip[0],gc->ip[1],gc->ip[2],gc->ip[3],gc->port);
		printf("pack: %i/%i/%i ",gc->sent_packets, gc->received_packets,gc->dropped_packets);
		printf("bytes: %i/%i\n", gc->sent_bytes, gc->received_bytes);
		printf("rate: %i/%i/%i ",gc->rate_sent_bytes, gc->rate_received_bytes, (int) gc->rate_start_time);
		printf("err: %i ",gc->routing_errors);
		printf("type: %i\n",gc->type);
		printf("state: %i  index: %i\n",gc->state,gc->index);
	} else {
		printf("tid: %li wasn't a prob\n",
#if defined(PTHREADS_DRAFT4) && defined(hpux)
			((HPUX_hack)gc->tid).longpart );
#else
			gc->tid);
#endif
	}
	return 0;
}		


void segv_handler(int a)
{
	
	segv_tid=pthread_self();
	
	fprintf(stderr,"Segfault!!! TID=%li  PID=%i\n",
#if defined(PTHREADS_DRAFT4) && defined(hpux)
		((HPUX_hack)pthread_self()).longpart,
#else
		(long) pthread_self(),
#endif
		getpid());
		
	fflush(stderr);
	sleep(1);
	fflush(stderr);
	// now I'm going to go through first the connection array and see if
	// this ID exists anywhere...
	
	gnut_connection_enumerate(segv_callback);
	gnut_transfer_enumerate(segv_callback2);	
	
#ifdef DMALLOC
	dmalloc_shutdown();
#endif
	
	printf("Waiting for gdb to work...\n");
	while(1) sleep(1);
	exit(0);
}

void reaper(int a)
{
	(void)wait(NULL);
}
#endif

int gnut_global_rate()
{
	return global_rate;
}

#ifdef PTHREADS_DRAFT4
pthread_once_t gnut_mutexes_once=pthread_once_init;
#endif

int gnut_init()
{
	char mac[6];
	char guid[16];
	char *d;
	int i;
#ifdef WIN32
	WORD wVersionRequested;
	WSADATA wsaData;
	int err;
	
	wVersionRequested = MAKEWORD( 2, 2);
	
	err=WSAStartup( wVersionRequested, &wsaData);
	if (err != 0) {
		return 0;
	}
#else	

	struct sigaction sa;

#ifdef PTHREADS_DRAFT4
	pthread_once(&gnut_mutexes_once, gnut_mutexes_init);
#endif

	memset(&sa,0,sizeof(sa));	

	sa.sa_handler=SIG_IGN;
	sa.sa_flags=SA_RESTART;

//	sigaction(2,&sa,NULL);
	
	sigaction(13,&sa,NULL);
	sigaction(28,&sa,NULL);
	sigaction(14,&sa,NULL);
	sigaction(15,&sa,NULL);
	sigaction(18,&sa,NULL);
	sigaction(19,&sa,NULL);
//	sigaction(20,&sa,NULL);
//	sigaction(21,&sa,NULL);
//	sigaction(22,&sa,NULL);
	sigaction(23,&sa,NULL);
	
	sigaction(10,&sa,NULL);
	sigaction(12,&sa,NULL);
	sigaction(29,&sa,NULL);
	sigaction(30,&sa,NULL);
	sigaction(6,&sa,NULL);
	
	signal(11,segv_handler);
	signal(17,reaper);
#endif
	
	srand(time(NULL));
	
	for (i=0;i<6;i++) mac[i]=rand();
//	conf_set_mac(mac);
	conf_set_str_len("mac",mac,6);
	
	for (i=0;i<6;i++) guid[i]=mac[i];
	for (i=7;i<16;i++) guid[i]=rand();
//	conf_set_guid(guid);
	conf_set_str_len("guid",guid,16);
	
	route_init();
	monitor_init();
	query_init();
	share_cache_init();
	host_init();
	
	d=expand_path("~/.gnut_hosts");
	g_debug(2,"trying to restore hosts\n");
	
	host_restore(d);
	
	free(d);
	
	return 0;
}


int gnut_transfer_start(query_response *qr)
{
	pthread_t nt;

#ifndef PTHREADS_DRAFT4
	pthread_create(&nt,NULL,gnut_threads_transfer,qr);
	pthread_detach(nt);
#else	
	pthread_create(&nt,pthread_attr_default,gnut_threads_transfer,qr);
	pthread_detach(&nt);
#endif

	return 0;
}	

// sets up up a new push connection...
int gnut_push_new(gnutella_push *gp)
{
	pthread_t np;
	push_arg *pa;
	share_item *si;
	int ref;
	
	memcpy(&ref,gp->ref,4);
	ref=GUINT32_FROM_LE(ref);
	
	si=share_find(ref);
	
	if (si==NULL) return -1;
	
	pa=(push_arg *) xmalloc(sizeof(push_arg));
	
	pa->si=si;
	memcpy(pa->ip,gp->ip,4);
	memcpy(pa->port,gp->port,2);

#ifndef PTHREADS_DRAFT4
	pthread_create(&np,NULL,gnut_threads_push,pa);
	pthread_detach(np);
#else	
	pthread_create(&np,pthread_attr_default,gnut_threads_push,pa);
	pthread_detach(&np);
#endif

	return 0;
}	

int gnut_try_host(void *a, void *b)
{
	host_entry *he;
	host_entry **he2;
	uint16 port;
	int flag;
	
	he=a;
	he2=b;
	
	memcpy(&port,he->port,2);
	port=GUINT16_FROM_LE(port);
	flag=gnut_connection_find(he->ip,port);
	if (memcmp(he->ip,net_local_ip(),4)==0) flag=0;
		else if (memcmp(he->ip,"\127\0\0\1",4)==0) flag=0;
		else if (memcmp(he->ip,"\0\0\0\0",4)==0) flag=0;
	
	if (flag!=0) {
		*he2=he;
		return -1;
	}	
	return 0;
}

// int gnut_outgoing_new(char *host)
//
// sets up a new outgoing host
//
int gnut_outgoing_new(char *host)
{
	char *mhost;
	pthread_t no;
	Gnut_Hash *gl;
	host_entry *he;
	uchar myip[4];
	uint16 mport;
	
	if (host!=NULL) {
		mhost=strdup(host);
	} else {
		// we need to get a host off the top of the host catcher,
		// in the process remove it...
		host_lock();
		gl=host_retrieve();
		
		he=NULL;
		gnut_hash_foreach(gl,gnut_try_host,&he);
		
		if (!he) {
			g_debug(4,"no eligible hosts in catcher\n");
			host_unlock();
			return 0;
		}
		memcpy(myip,he->ip,4);
		memcpy(&mport,he->port,2);
		mport=GUINT16_FROM_LE(mport);
		host_unlock();
		mhost=(char *) xmalloc(100);
		sprintf(mhost,"%i.%i.%i.%i:%i",myip[0],myip[1],myip[2],myip[3],mport);
	}
			
#ifndef PTHREADS_DRAFT4
	pthread_create(&no,NULL,gnut_threads_outgoing,mhost);
	pthread_detach(no);
#else
	pthread_create(&no,pthread_attr_default,gnut_threads_outgoing,mhost);
	pthread_detach(&no);
#endif
	
	return 0;
}

int gnut_incoming_new(int sock)
{
	int newsock;
	struct sockaddr_in sin;
	int len;
	pthread_t ni;
	incoming_arg *ia;
	
	g_debug(2,"entering\n");
	
	len=sizeof(sin);
	newsock=accept(sock,(struct sockaddr *) &sin,&len);
	
	if (newsock<0) {
		perror("gnut_incoming_new, accept");
		return newsock;
	}

	ia=(incoming_arg *)xmalloc(sizeof(incoming_arg));
	ia->s=newsock;
	memcpy(&ia->sin,&sin,sizeof(sin));

#ifndef PTHREADS_DRAFT4
	pthread_create(&ni,NULL,gnut_threads_incoming,ia);
	pthread_detach(ni);
#else
	pthread_create(&ni,pthread_attr_default,gnut_threads_incoming,ia);
	pthread_detach(&ni);
#endif
	
	g_debug(2,"leaving\n");
	return 0;
}

int count_rate(gnut_transfer *gt)
{
	float tr_rate;
	float a;
	time_t t;
	
	if ((gt->state&1)==0 || gt->state==STATE_DEAD || 
		gt->state==STATE_ERROR || gt->state==STATE_CLEANABLE) 
		return 0;
	
	a=(float)gt->rate_bytes;
	t=time(NULL) - gt->rate_start_time;
	if (t==0) {
		tr_rate=0;
	} else {
		tr_rate=a / (float) t;
	}
	g_debug(3,"tr_rate=%f new_global_rate=%f t=%i a=%f\n",
		tr_rate,new_global_rate,(int) t, a);
	new_global_rate += (int) tr_rate;
	return 0;
}
	

void *main_loop(void *arg)
{
	int listen_socket;
	int port,ret;
	fd_set fsr;
	struct timeval tv;

	g_debug(3,"entering\n");
	
//	port=conf_get_local_port();
	port=conf_get_int("local_port");
	listen_socket=create_listen_socket(&port);
	if (listen_socket<0) return NULL;
	
//	conf_set_local_port(port);
	conf_set_int("local_port",port);
	
	while (1) {
		FD_ZERO(&fsr);
		FD_SET(listen_socket,&fsr);
		
		tv.tv_sec=0;
		tv.tv_usec=250000;
		
		ret=select(listen_socket+1,&fsr,NULL,NULL,&tv);
		g_debug(5,"select returned: %i\n",ret);
		
		if (FD_ISSET(listen_socket,&fsr)) {
			g_debug(2,"incoming connection\n");
			gnut_incoming_new(listen_socket);
		}
		
		if (conf_get_int("min_connections") > (gnut_threads_num_incoming() + gnut_threads_num_outgoing()) ) {	
			g_debug(4,"initiating outgoing\n");
			gnut_outgoing_new(NULL);
		}
		
		// now we calculate the global transfer rate...
		new_global_rate=0;
		gnut_transfer_enumerate(count_rate);
		global_rate=new_global_rate;
	}
	
	return NULL;
}

int gnut_start_main_loop()
{
	pthread_t ml;

#ifndef PTHREADS_DRAFT4
	pthread_create(&ml,NULL,main_loop,NULL);
	pthread_detach(ml);
#else	
	pthread_create(&ml,pthread_attr_default,main_loop,NULL);
	pthread_detach(&ml);
#endif

	return 0;
}
