/* Josh Pieper, (c) 2000 */

/* This file is distributed under the GPL, see file COPYING for details */

#include <stdio.h>
#include <stdlib.h>

#ifndef WIN32
#include <unistd.h>
#else
#include <io.h>
#endif

#include <errno.h>
#include <pthread.h>

#include "gnut_lib.h"
#include "gnut_net.h"
#include "gnut_threads.h"
#include "gnut_transfer.h"
#include "Gnut_List.h"
#include "gnut_connection.h"
#include "conf.h"
#include "gnut.h"

#ifndef PTHREADS_DRAFT4
pthread_mutex_t transfer_mutex=PTHREAD_MUTEX_INITIALIZER;
#else
pthread_mutex_t transfer_mutex;
#endif

Gnut_List *transfer_list=NULL;

int gnut_transfer_num()
{
	return gnut_list_size(transfer_list);
}

// this is called after the file has been opened and the socket
// set up, it either sends or receives the entire content of the
// file, updating the gnut_transfer struc along the way...
// It leaves the structure afterwards though, so that the user
// can see downloads that occurred while he/she was away.
int gnut_transfer_loop(gnut_transfer *gt)
{
	char *netbuf;
	int readfd,writefd;
	int ret=0,ret2,left;
	int rate=0;
	int i;
	int download;
	
	g_debug(3,"entering gt=%p\n",gt);
	
	if ( (gt->type & 1) == 0 ) {
		// this is a download
		g_debug(3,"download!\n");
		readfd=gt->sock;
		writefd=gt->fsock;
		download=1;
	} else {
		// this is an upload
		g_debug(3,"upload!\n");
		readfd=gt->fsock;
		writefd=gt->sock;
		download=0;
	}

	netbuf=(char *)xmalloc(4096);
	left=gt->total - gt->bytes;
	g_debug(3,"entering loop\n");
#ifdef WIN32
#undef close
#undef read
#undef write
#endif
	while (gt->state==STATE_CONNECTED && left>0) {
		do {
			// this recv crap is here because win32 doesn't treat
			// socks and fd's the same.... :(
			if (download) ret=timed_recv(readfd,netbuf,MIN(left,4096),0,300);
				else ret=read(readfd,netbuf,MIN(left,4096));
		} while (errno==EINTR && ret<0);
		if (ret<0) {
			if (gnut_lib_debug>2) perror("gnut_transfer_loop, read");
		}
		if (ret==0) break;
		if (ret>0) {
			do {
				if (download) ret2=write(writefd,netbuf,ret);
					else ret2=send(writefd,netbuf,ret,0);
			} while (errno==EINTR && ret2<0);
			if (ret2<=0) break;
			g_debug(5,"copied %i bytes from sock to sock\n",ret);
			gt->bytes+=ret;
			gt->rate_bytes+=ret;
			left-=ret;
		} else break;
		if (gt->rate_limit!=0 || conf_get_int("global_bandwidth_cap")!=0) {
			// we have a rate limit, and we need to see if we're
			// exceeding it
			g_debug(3,"entering rate limiting code\n");
			do {
				g_debug(3,"rate limiting loop.. rate=%i\n",rate);
				g_debug(3,"gnut_global_rate: %i\n",gnut_global_rate());
				i=time(NULL) - gt->rate_start_time;
				if (i==0) i=1;
				rate=gt->rate_bytes/i;

//				printf("%i/%i=%i\n",gt->rate_bytes,i,rate);
				if (gt->rate_limit!=0 && rate>gt->rate_limit) sleep(1);
				if (conf_get_int("global_bandwidth_cap")!=0 &&
					gnut_global_rate()>conf_get_int("global_bandwidth_cap")) sleep(1);
			} while ( ( gt->rate_limit!=0 && rate>gt->rate_limit ) || 
					(conf_get_int("global_bandwidth_cap")!=0 && 
						gnut_global_rate()>conf_get_int("global_bandwidth_cap") ));
			g_debug(3,"my rate is: %i and I want to be: %i\n",rate,gt->rate_limit);
		}
	}
	
	
	g_debug(3,"closing connection on ret: %i\n",ret);
	free(netbuf);
	
	if (download) shutdown(readfd,2);
		shutdown(writefd,2);
	close(writefd);
	close(readfd);
	gt->sock=-1;
	gt->fsock=-1;
	
	if (left==0) {
		g_debug(3,"returning success\n");
		return 0;	
	} else {
		g_debug(3,"returning partial\n");
		return -1;
	}		
}

gnut_transfer * gnut_transfer_new()
{
	gnut_transfer *gt;
	
	g_debug(3,"entering\n");
	
	gt=(gnut_transfer *) xmalloc(sizeof(gnut_transfer));
	
	gt->sock=-1;
	gt->fsock=-1;
	gt->fname=NULL;
	
	gt->type=TYPE_UNSPEC;
	gt->bytes=0;
	gt->total=0;
	
	gt->rate_bytes=0;
	gt->rate_start_time=time(NULL);
	gt->rate_limit=0;
	
	gt->state=STATE_UNCON;
	
	gt->tid=pthread_self();
	
	pthread_mutex_lock(&transfer_mutex);
	
	transfer_list=gnut_list_prepend(transfer_list,gt);
	
	pthread_mutex_unlock(&transfer_mutex);
	
	g_debug(3,"returning success\n");
	return gt;
}
	
int gnut_transfer_delete(gnut_transfer *gt)
{
	g_debug(3,"entering\n");
	
	pthread_mutex_lock(&transfer_mutex);
	
	transfer_list=gnut_list_remove(transfer_list,gt);
	
	pthread_mutex_unlock(&transfer_mutex);
	
	if (gt->fname) free(gt->fname);
	
	free(gt);
	
	g_debug(3,"returning success\n");
	
	return 0;
}

int gnut_transfer_enumerate( int(*a)(gnut_transfer *))
{
	Gnut_List *tmp;

	pthread_mutex_lock(&transfer_mutex);
	
	for (tmp=transfer_list;tmp;tmp=tmp->next) {
		if ((*a)(tmp->data)==-1) break;
	}
	
	pthread_mutex_unlock(&transfer_mutex);
	return 0;
}

	