/* Josh Pieper, (c) 2000 */

/* This file is distributed under the GPL, see file COPYING for details */

#include <stdio.h>
#include <stdlib.h>
#include <pthread.h>

#include "gnut_lib.h"
#include "Gnut_Hash.h"
#include "Gnut_Queue.h"
#include "route.h"

// this mutex protects all finds, adds, and removes to the
// routing table
#ifndef PTHREADS_DRAFT4
pthread_mutex_t route_mutex=PTHREAD_MUTEX_INITIALIZER;
#else
pthread_mutex_t route_mutex;
#endif

Gnut_Hash *route_hash=NULL;
Gnut_Queue *route_queue=NULL;

int route_compare_func(void *a, void *b)
{
	gnut_mprobe(a);
	gnut_mprobe(b);
	if (memcmp(a,b,16)==0) return 0;
	return -1;
}

uchar route_hash_func(void *a)
{
	route_entry *re;
	uchar b;
	int i;
	
//	g_debug(1,"entering a=%p\n",a);
	
	re=a;
	for (b=0,i=0;i<16;i++) {
//		printf("%02x",re->guid[i]);
		b+=re->guid[i];
	}
	
//	printf("\n");
	return b;
}

int route_init()
{
	route_hash=gnut_hash_new(route_hash_func,route_compare_func);
	route_queue=gnut_queue_new();
	return 0;
}

// int route_guid_add(uchar guid[16], gnut_connection *gc)
//
// adds the guid (which came over connection gc) into the routing table
//
// This needs to limit the size by removing the old ones.
//
int route_guid_add(uchar guid[16], gnut_connection *gc)
{
	route_entry *re;
	
	g_debug(3,"entering gc=%p\n",gc);
		
	re=(route_entry *) xmalloc(sizeof(route_entry));
	
	memcpy(re->guid,guid,16);
	re->p=gc;
	
	pthread_mutex_lock(&route_mutex);

//	printf("route_guid_add: ");
//	for (i=0;i<16;i++) printf("%02X",re->guid[i]);

//	printf("\n\n");	
	gnut_hash_insert(route_hash,re);
	gnut_queue_insert(route_queue,re);
	
	if (route_queue->size>ROUTE_MAX) {
		g_debug(5,"hit max routing table size\n");
		re=gnut_queue_remove(route_queue);
		gnut_hash_remove(route_hash,re);
		xfree(re);
	}

	
	pthread_mutex_unlock(&route_mutex);
	
	g_debug(3,"returning success\n");
	return 0;
}

// gnut_connection * route_guid_find(uchar guid[16])
//
// looks through the routing table, and if an entry is found,
// the corresponding connection is returned, otherwise
// NULL is returned
//
gnut_connection *route_guid_find(uchar guid[16])
{
	route_entry *re,*re2;
	
	g_debug(3,"entering\n");
	
	re=(route_entry *) xmalloc(sizeof(route_entry));
	gnut_mprobe(re);

	memcpy(re->guid,guid,16);
	
	pthread_mutex_lock(&route_mutex);
	g_debug(3,"locked mutex\n");
	
	re2=gnut_hash_find(route_hash,re);
		
	pthread_mutex_unlock(&route_mutex);
	
	gnut_mprobe(re);
	
	xfree(re);
	if (re2) return re2->p;
	g_debug(3,"not found!\n");
	return NULL;
}
			
// int route_guid_clear(gnut_connection *gc)
//
// turns all occurances of gc in routing table into NULL
// so that they won't be matched again
//
int route_guid_clear(gnut_connection *gc)
{
	Gnut_List *tmp;
	route_entry *re;
	g_debug(4,"entering gc=%p",gc);
	
	pthread_mutex_lock(&route_mutex);
	g_debug(4,"locked mutex\n");
	
	// since the hash contains pointers to the
	// same data, I can do this here, and save a lot
	// of time...
	for (tmp=gnut_queue_list(route_queue);tmp;tmp=tmp->next) {
		gnut_mprobe(tmp);
		re=tmp->data;
		if (re->p==gc) re->p=NULL;
	}
	
	pthread_mutex_unlock(&route_mutex);
	
	g_debug(4,"returning success\n");
	return 0;
}
	

			