/* Josh Pieper, (c) 2000 */

/* This file is distributed under the GPL, see file COPYING for details */

#include <stdio.h>
#include <stdlib.h>
#include <sys/stat.h>

#ifndef WIN32
#include <unistd.h>
#endif

#include <errno.h>
#include <string.h>
#include <pthread.h>
#include <ctype.h>

#include "gnut_lib.h"
#include "share.h"
#include "Gnut_List.h"


// here we declare share's global variables...
int share_num=0;
int share_size=0;
Gnut_List *share_root=NULL;

#ifndef PTHREADS_DRAFT4
pthread_mutex_t share_mutex=PTHREAD_MUTEX_INITIALIZER;
#else
pthread_mutex_t share_mutex;
#endif

// called by share_search to check a query...
int is_match_tok(char *query, char *str)
{
	char *a,*b;
	
	g_debug(5,"entering\n");

#ifdef HAVE_STRTOK_R	
	a=strtok_r(query," \t",&b);
#else
	a=strtok(query," \t");
#endif

	if (a==NULL) return 0;
	
	while (a!=NULL) {
		g_debug(5,"checking for \"%s\" in \"%s\" \n",a,str);
		if (strstr(str,a)==NULL) {
			g_debug(5,"returning failure\n");
			return 0;
		}
#ifdef HAVE_STRTOK_R
		a=strtok_r(NULL," \t",&b);
#else
		a=strtok(NULL," \t");
#endif
	}
	
	g_debug(5,"returning success\n");
	return 1;
}

// Gnut_List *share_search(char *query, int max)
//
// searches through the shares, and returns a list of up to <max>
// shares that match the query
//
Gnut_List *share_search(char *orig_query, int max)
{
	Gnut_List *nlist=NULL,*ltmp;
	share_item *si;
	share_item *copy;
	char *myquery,*query;
	int count;
	int i,len;
	
	g_debug(3,"entering\n");
	
	query=strdup(orig_query);
	gnut_strdelimit(query,"*?._<>|[]{}()_=+-/!@#$%^&",' ');
	
	len=strlen(query);
	for (i=0;i<len;i++) query[i]=tolower(query[i]);
	
	count=0;
	myquery=strdup(query);
	
	for (ltmp=share_root;ltmp!=NULL;ltmp=ltmp->next) {
		g_debug(4,"ltmp=%p myquery=%p query=%p\n",ltmp,myquery,query);
		// see if this share matches...
		si=ltmp->data;
		strcpy(myquery,query);
		
		gnut_mprobe(ltmp);
		gnut_mprobe(myquery);
		gnut_mprobe(query);
		
		if (is_match_tok(myquery,si->upath)) {
			// we need to make a copy of this node, and
			// insert into the result list...
			
			copy=(share_item *) xmalloc(sizeof(share_item));
			
			copy->path=strdup(si->path);
			copy->upath=strdup(si->upath);
			copy->fpath=strdup(si->fpath);
			copy->size=si->size;
			copy->ref=si->ref;
			
			nlist=gnut_list_prepend(nlist,copy);
			
			if ((++count)>max) break;
		}
	}
	xfree(myquery);
	xfree(query);
	
	g_debug(3,"returning\n");
	return nlist;	
} 

// Gnut_List *share_get_root()
//
// returns the share list, used to do ui
//
Gnut_List *share_get_root()
{
	return share_root;
}

// int share_totals(int *num, int *size)
//
// returns total number of shared files, and their total size
// in KB
//
int share_totals(int *num, int *size)
{
	g_debug(3,"entering\n");
	*num=share_num;
	*size=share_size;
	g_debug(3,"returning share_num=%i share_size=%i *num=%i *size=%i\n",share_num,share_size,*num,*size);
	return 0;
}

int share_insert(void *data, void *user_data)
{
	share_item *si,*nsi;
	Gnut_List **l;
	
	
	l=user_data;
	
	si=data;
	nsi=(share_item *) xmalloc(sizeof(share_item));
	memcpy(nsi,si,sizeof(share_item));
	nsi->path=strdup(si->path);
	nsi->upath=strdup(si->upath);
	nsi->fpath=strdup(si->fpath);
	
	*l=gnut_list_prepend(*l,nsi);
	return 0;
}

Gnut_List * share_copy_list(Gnut_List *l)
{
	Gnut_List *nl=NULL;
	g_debug(3,"entering\n");
	
	gnut_list_foreach(l,share_insert,&nl);
	
	return nl;
}

// callback, free's up path in share_item
int share_delete(void * data, void * user_data)
{
	share_item *si;

	g_debug(3,"entering\n");

	si=data;
	xfree(si->path);
	xfree(si->upath);
	xfree(si->fpath);
	g_debug(3,"returning\n");
	return 0;
}

// int share_clear_list(Gnut_List *l)
//
// clears the share list specified by l
//
int share_clear_list(Gnut_List *l)
{
	g_debug(3,"entering\n");

	gnut_list_foreach(l,share_delete,NULL);
	gnut_list_free(l);

	g_debug(3,"returning\n");
	return 0;
}
	
	
// int share_clear()
//
// clears out the list of shares
//
int share_clear()
{
	g_debug(3,"entering\n");
	gnut_list_foreach(share_root,share_delete,NULL);
	gnut_list_free(share_root);
	share_root=NULL;
	share_num=0;
	share_size=0;
	g_debug(3,"returning\n");
	return 0;
}		
	
// share_item *share_find(int ref)
//
// locates the share indexed by ref and returns it
//
share_item *share_find(int ref)
{
	Gnut_List *ltmp;
	share_item *si,*si2;
	
	g_debug(3,"entering\n");
	
	for (ltmp=share_root;ltmp!=NULL;ltmp=gnut_list_next(ltmp)) {
		si=ltmp->data;
		if (si->ref==ref) {
			// we found the correct one, now we'll make a copy
			// of it and return...
			si2=(share_item *) xmalloc(sizeof(share_item));
			si2->ref=si->ref;
			si2->size=si->size;
			si2->path=strdup(si->path);
			si2->upath=strdup(si->upath);
			si2->fpath=strdup(si->fpath);
			g_debug(3,"returning si2=%p\n",si2);
			return si2;
		}
	}
	g_debug(3,"returning failure\n");
	return NULL;
}

#ifdef WIN32
void log_win32_error(char *szFunctionName)
{
	LPVOID lpMsgBuf;
	FormatMessage(
		FORMAT_MESSAGE_ALLOCATE_BUFFER |
		FORMAT_MESSAGE_FROM_SYSTEM |
		FORMAT_MESSAGE_IGNORE_INSERTS,
		NULL,
		GetLastError(),
		MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT), 
		(LPSTR) &lpMsgBuf,
		0,
		NULL
		);
		
	g_debug(3,"%s: %s\n",szFunctionName, (LPSTR) lpMsgBuf);
	LocalFree(lpMsgBuf);
}

// copies into pszFilePath a version of szPath that is ensured to end in a "\*.*"

void cleanup_directory_path(char *szPath, char *pszFilePath, int *pnFilePathLength)
{
	strcpy(pszFilePath, szPath);
	*pnFilePathLength = strlen(szPath);
	
	// TODO: if szPath ends in \*.* bail out
	
	if (L'\\' != *( pszFilePath + *pnFilePathLength-1)) {
		strcat(pszFilePath, TEXT("\\"));
	}
		
	strcat(pszFilePath, TEXT("*.*"));
}

int share_path_len=0;

int share_scan_dir_win32(char *szPath)
{
	char szFilePath[OUR_MAX_PATH+1];
	char szRootFilePath[OUR_MAX_PATH+1];
	ULONG	nFilePathLength;
	DWORD	dwFileAttributes;
	HANDLE	hSearch;
	WIN32_FIND_DATA	Win32FindData;
	share_item *si;
	int i, len;
	
	dwFileAttributes = GetFileAttributesA( szPath );
	if (0xFFFFFFFF == dwFileAttributes)
		return 0;
	
	if (dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY) {
		cleanup_directory_path(szPath, szRootFilePath, &nFilePathLength);
	}
	
	hSearch = FindFirstFile (szRootFilePath, &Win32FindData);
	if (INVALID_HANDLE_VALUE == hSearch) {
		log_win32_error("share_scan_dir_win32");
		return 0;
	}
	
	// Make sure we ignore the "." and ".." files
	while (!strcmp(Win32FindData.cFileName, TEXT("."))
		|| !strcmp(Win32FindData.cFileName, TEXT("..")))
	{
		if (!FindNextFile(hSearch, &Win32FindData)) {
			log_win32_error("share_scan_dir_win32");
			FindClose(hSearch);
			return 0;
		}
	}
	
	do {
		strcpy(szFilePath, szPath);
		nFilePathLength = strlen( szFilePath );
		if (L'\\' != *(szFilePath + nFilePathLength - 1)) {
			strcat(szFilePath, TEXT("\\"));
		}
		strcat(szFilePath, Win32FindData.cFileName);
		
		if (Win32FindData.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY) {
			g_debug(3, "share_scan_dir_win32: scanning %s \n",szFilePath);
			share_scan_dir_win32(szFilePath);
		} else {
			si = (share_item *)xmalloc(sizeof(share_item));
			si->size=Win32FindData.nFileSizeLow;
			
			share_size+= (si->size/1024);
			si->ref=++share_num;
			si->path=strdup(szFilePath+share_path_len+1);
			gnut_strdelimit(si->path,path_slash,'/');
			si->upath=strdup(szFilePath+share_path_len+1);
			gnut_strdelimit(si->upath,path_slash,'/');
			si->fpath=strdup(szFilePath);
			
			len=strlen(si->upath);
			
			// lowercase the path
			for (i=0;i<len;i++) si->upath[i]=tolower(si->upath[i]);
			
			pthread_mutex_lock(&share_mutex);
			share_root=gnut_list_prepend(share_root,si);
			pthread_mutex_unlock(&share_mutex);
		}
	} while (FindNextFile(hSearch, &Win32FindData));

	FindClose(hSearch);
	return 0;
}

#endif

// int scan_dir(char *path)
//
// adds the contents of path to the share_root list
//
int share_scan_dir(char *path)
{
#ifdef WIN32
	share_path_len=strlen(path);
	return share_scan_dir_win32(path);
#else
	FILE *findfp;
	char *buf;
	struct stat st;
	int ret;
	share_item *si;
	int i,len,pathlen;
	
	g_debug(3,"entering\n");
	
	pthread_mutex_lock(&share_mutex);
	
	pathlen=strlen(path);
	buf=(char *) xmalloc(pathlen+OUR_MAX_PATH);
	sprintf(buf,"find %s -follow \\( -type f -o -type l \\) -print",path);
	
	g_debug(1,"about to run: \"%s\" \n",buf);
	
	findfp=popen(buf,"r");
	if (findfp==NULL) {
		g_debug(1,"pipe to find returned error: %i\n",errno);
		pthread_mutex_unlock(&share_mutex);
		return -1;
	}
	g_debug(1,"retrieving results from find...\n");
	while (fgets(buf,256,findfp)!=NULL) {
		buf[strlen(buf)-1]=0;	// remove trailing eol
		
		ret=stat(buf,&st);
		
		si=(share_item *) xmalloc(sizeof(share_item));
		
		si->size=st.st_size;
		si->mtime=st.st_mtime;
		share_size+=(st.st_size/1024);
		si->ref=++share_num;
		
		si->path=strdup(buf+pathlen+1);
		si->upath=strdup(buf+pathlen+1);
		si->fpath=strdup(buf);
		len=strlen(si->upath);
		for (i=0;i<len;i++) si->upath[i]=tolower(si->upath[i]);
		
		share_root=gnut_list_prepend(share_root,si);
	}
	
	pclose(findfp);
	xfree(buf);
	
	pthread_mutex_unlock(&share_mutex);
	
	return 0;
#endif
}
		
		
